"""ContactSelf"""

from __future__ import annotations
from typing import Any, Optional, Type

from wechaty import FileBox, get_logger
from wechaty.exceptions import WechatyOperationError
from wechaty.user.contact import Contact

log = get_logger('ContactSelf')


class ContactSelf(Contact):
    """ContactSelf"""

    async def avatar(self, file_box: Optional[FileBox] = None) -> FileBox:
        """get avatar of ContactSelf

        Args:
            file_box (Optional[FileBox], optional): i. Defaults to None.

        Examples:
            >>> contact_self = bot.contact_self()
            >>> file_box = await contact_self.avatar()

        Raises:
            WechatyOperationError: _description_

        Returns:
            FileBox: _description_
        """
        log.info('avatar(%s)' % file_box.name if file_box else '')
        if not file_box:
            file_box = await super().avatar(None)
            return file_box

        if self.contact_id != self.puppet.self_id():
            raise WechatyOperationError('set avatar only available for user self')

        await self.puppet.contact_avatar(self.contact_id, file_box)

    async def qr_code(self) -> str:
        """return the qrcode of ContactSelf

        Raises:
            WechatyOperationError: if there is login exception, it will not get the qrcode
            WechatyOperationError: if the contact is not self, it will not get the qrcode

        Returns:
            str: the content of qrcode
        """
        try:
            contact_id: str = self.puppet.self_id()
        except Exception:
            raise WechatyOperationError(
                'Can not get qr_code, user might be either not logged in or already logged out'
            )

        if self.contact_id != contact_id:
            raise WechatyOperationError('only can get qr_code for the login user self')
        qr_code_value = await self.puppet.contact_self_qr_code()
        return qr_code_value

    @property
    def name(self) -> str:
        """
        get the name of login contact

        Examples:
            >>> contact_self = bot.contact_self()
            >>> name = contact_self.name

        Returns:
            the name of Login User
        """
        return super().name

    async def set_name(self, name: str) -> None:
        """
        set the name of login contact

        Args:
            name: new name
        """
        await self.puppet.contact_self_name(name)
        await self.ready(force_sync=True)

    async def signature(self, signature: str) -> Any:
        """

        :param signature:
        :return:
        """
        puppet_id = self.puppet.self_id()

        if self.contact_id != puppet_id:
            raise WechatyOperationError('only can get qr_code for the login user self')

        return self.puppet.contact_signature(signature)



"""
UrlLink for Contact Message
"""
from __future__ import annotations

from typing import (
    Optional,
    Type
)
from wechaty_puppet import UrlLinkPayload, get_logger

from wechaty.utils.link import get_url_metadata



log = get_logger('UrlLink')


class UrlLink:
    """
    url_link object which handle the url_link content
    """

    def __init__(
        self,
        payload: UrlLinkPayload,
    ):
        """
        initialization
        :param payload:
        """
        self.payload: UrlLinkPayload = payload

    @classmethod
    def create(
        cls: Type[UrlLink],
        url: str,
        title: Optional[str] = None,
        thumbnail_url: Optional[str] = None,
        description: Optional[str] = None
    ) -> UrlLink:
        """
        create urllink from url string
        """
        log.info('create url_link for %s', url)

        metadata = get_url_metadata(url)

        payload = UrlLinkPayload(url=url)
     
        payload.title = title or metadata.get('title', None)
        payload.thumbnailUrl = thumbnail_url or metadata.get('image', None)
        payload.description = description or metadata.get('description', None)
        return UrlLink(payload)

    def __str__(self) -> str:
        """
        UrlLink string format output
        :return:
        """
        return 'UrlLink<%s>' % self.payload.url

    @property
    def title(self) -> str:
        """
        get UrlLink title
        :return:
        """
        return self.payload.title or ''

    @property
    def thumbnailUrl(self) -> str:
        """
        get thumbnail url
        :return:
        """
        return self.payload.thumbnailUrl or ''

    @property
    def description(self) -> str:
        """
        get description
        :return:
        """
        return self.payload.description or ''

    @property
    def url(self) -> str:
        """
        get url
        :return:
        """
        return self.payload.url or ''



"""
async helpers
"""
from __future__ import annotations
import asyncio
from asyncio import Task
from typing import (
    List,
    Any,
    Optional,
    Set,
    TYPE_CHECKING
)
import functools

if TYPE_CHECKING:
    from wechaty import Message, WechatyPlugin


async def gather_with_concurrency(n_task: int, tasks: List[Task]) -> Any:
    """
    gather tasks with the specific number concurrency
    Args:
        n_task: the number of tasks
        tasks: task objects
    """
    semaphore = asyncio.Semaphore(n_task)

    async def sem_task(task: Task) -> Any:
        async with semaphore:
            return await task
    return await asyncio.gather(*(sem_task(task) for task in tasks))


class SingleIdContainer:
    """Store the Message Id Container"""
    _instance: Optional[SingleIdContainer] = None

    def __init__(self) -> None:
        self.ids: Set[str] = set()
        self.max_size: int = 100000

    def exist(self, message_id: str) -> bool:
        """exist if the message has been emitted

        Args:
            message_id (str): the identifier of message

        Returns:
            bool: if the message is the first message
        """
        if message_id in self.ids:
            return True
        self.ids.add(message_id)
        return False
    
    @classmethod
    def instance(cls) -> SingleIdContainer:
        """singleton pattern for MessageIdContainer"""
        if cls._instance is None or len(cls._instance.ids) > cls._instance.max_size:
            cls._instance = SingleIdContainer()

        return cls._instance


def single_message(on_message_func):    # type: ignore
    """single message decorator"""
    @functools.wraps(on_message_func)
    async def wrapper(plugin: WechatyPlugin, message: Message) -> None:
        if not SingleIdContainer.instance().exist(message.message_id):
            await on_message_func(plugin, message)
    return wrapper
